// Probabilistic reachability set storage -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "HashGraph.h"
#include "ByteBuffer.h"
#include <cstdlib>
#include <cassert>
#include <ctime>

#if defined __SUNPRO_CC
using std::calloc;
using std::free;
using std::rand;
using std::srand;
#endif // __SUNPRO_CC

/** @file HashGraph.C
 * Transient, probabilistic reachability set storage
 */

/** Number of bits size_t holds */
#define SIZE_T_BIT (sizeof (size_t) * CHAR_BIT)

/** primes[x] == prime closest to 32 << x */
static unsigned
primes[] = {
  0x1f,
  0x3d,
  0x7f,
  0x101,
  0x1fd,
  0x3fd,
  0x805,
  0xffd,
  0x1fff,
  0x3ffd,
  0x8003,
  0xfff1,
  0x1ffff,
  0x3fffb,
  0x7ffff,
  0xffffd,
  0x1ffff7,
  0x3ffffd,
  0x7ffff1,
  0xfffffd,
  0x2000023,
  0x3fffffb,
  0x800001d,
  0x10000003,
  0x1ffffffd,
  0x40000003,
  0x7fffffff,
  0xfffffffb
};

HashGraph::~HashGraph ()
{
  for (unsigned s = myNumTables; s--; ) {
    free (myHashes[s]);
    free (myHashers[s]);
  }
  free (myHashes);
  free (myHashers);
}

bool
HashGraph::init (unsigned numTables,
		 unsigned& funcSize,
		 unsigned& hashSize)
{
  assert (!myHashes && !myHashers && !myNumTables && !getNumStates ());
  if (!openFile ())
    return false;
  unsigned s;
  /* round the hash function size to the next power of 2 */
  for (s = 2; s < funcSize && sizeof (size_t) * s << 1; s <<= 1);
  funcSize = s;
  /* round the hash table size to the closest prime in our table */
  for (s = 0; s < ((sizeof primes) / sizeof *primes) - 1; s++)
    if (primes[s] >= hashSize)
      break;
  hashSize = primes[s];
  s = 1 + (hashSize - 1) / SIZE_T_BIT;

  /* allocate the tables */
  if (!(myHashers =
	static_cast<size_t**>(calloc (numTables, sizeof *myHashers))) ||
      !(myHashes =
	static_cast<size_t**>(calloc (numTables, sizeof *myHashes)))) {
  failure:
    if (myHashers) free (myHashers), myHashers = 0;
    if (myHashes) free (myHashes), myHashes = 0;
    return false;
  }

  unsigned t;
  for (t = numTables; t--; ) {
    if (!(myHashers[t] =
	  static_cast<size_t*>(calloc (funcSize, sizeof *myHashers[t]))) ||
	!(myHashes[t] =
	  static_cast<size_t*>(calloc (s, sizeof *myHashes[t])))) {
      for (; t < numTables; t++)
	free (myHashers[t]), free (myHashes[t]);
      goto failure;
    }
  }

  /* Initialize the tables for the hash functions */
  srand (time (0));
  for (t = numTables; t--; )
    for (s = funcSize; s--; )
      myHashers[t][s] =	size_t ((hashSize + double (1)) * rand () /
				(RAND_MAX + double (1)));
  myNumTables = numTables;
  myFuncSize = funcSize - 1;
  myHashSize = hashSize;
  return true;
}

bool
HashGraph::do_add (const void* buf,
		   size_t size)
{
  /** flag: was the state found? */
  bool found = true;
  /** the encoded byte stream */
  const char* state = reinterpret_cast<const char*>(buf);

  for (unsigned t = myNumTables; t--; ) {
    /* compute the hash value */
    size_t h = 0;
    const size_t* hasher = myHashers[t];
    for (size_t s = size; s--; )
      h += state[s] * hasher[s & myFuncSize];
    /* test and set the corresponding bit in the hash table */
    h %= myHashSize;
    size_t& ht = myHashes[t][h / SIZE_T_BIT];
    size_t mask = size_t (1) << (h % SIZE_T_BIT);
    if (!(ht & mask))
      ht |= mask, found = false;
  }

  if (found)
    return false;

  newState ();
  mySearch.push (buf, size, ftell (myPathFile));
  class BytePacker p;
  p.append (myOffset), p.append (size);
  fwrite (p.getBuf (), 1, p.getLength (), myPathFile);
  fwrite (buf, 1, size, myPathFile);
  return true;
}

word_t*
HashGraph::getState (long pos) const
{
  unsigned char rbuf[8];
  class ByteUnpacker u (rbuf);
  fseek (myPathFile, pos, SEEK_SET);
  fread (rbuf, sizeof rbuf, 1, myPathFile);
  unsigned offset = u.extract ();
  unsigned size = u.extract (); u.buf = rbuf;
  word_t* state = new word_t[size + (sizeof (word_t) - 1) / sizeof (word_t)];
  fseek (myPathFile,
	 pos + BytePacker::size (offset) + BytePacker::size (size), SEEK_SET);
  fread (state, 1, size, myPathFile);
  BitPacker::inflate (state[(size - 1) / sizeof (word_t)],
		      (-size) % sizeof (word_t));
  return state;
}

word_t*
HashGraph::pop (bool tail, size_t* size)
{
  size_t numBytes;
  word_t* buf;
  if (const size_t num = mySuppressed.size ()) {
    const struct StateSet::States::state& s = mySuppressed[num - 1];
    switch (mySuccessors) {
    case One:
      numBytes = s.size;
      buf = new word_t[1 + (numBytes - 1) / sizeof (word_t)];
      memcpy (buf, s.buf, numBytes);
      mySuccessors = None;
      goto fetch;
    case Many:
      const_cast<class HashGraph*>(this)->do_add (s.buf, s.size);
      mySuccessors = None;
      // fall through
    case None:
      mySuppressed.clear ();
    }
  }

  if (mySearch.empty ())
    return 0;

  buf = mySearch.pop (tail, myOffset, &numBytes);

  if (compress && !mySuppressed.add (buf, numBytes))
    assert (false);

 fetch:
  if (size)
    *size = numBytes;
  else
    BitPacker::inflate (buf[(numBytes - 1) / sizeof (word_t)],
			(-numBytes) % sizeof (word_t));
  return buf;
}
