// Probabilistic reachability set storage -*- c++ -*-

#ifndef HASHGRAPH_H_
# define HASHGRAPH_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include "StateSet.h"

/** @file HashGraph.h
 * Transient, probabilistic reachability set storage
 */

/** Probabilistic reachability set storage */
class HashGraph : public StateSet
{
public:
  /** Constructor
   * @param compress	flag: apply path compression
   */
  HashGraph (bool compress) : StateSet (compress),
    myNumTables (0), myFuncSize (0), myHashSize (0),
    myHashers (0), myHashes (0) {}
private:
  /** Copy constructor */
  HashGraph (const class HashGraph& old);
  /** Assignment operator */
  class HashGraph& operator= (const class HashGraph& old);
public:
  /** Destructor */
  ~HashGraph ();

  /** Initialize the hash tables
   * @param numTables	number of hash tables
   * @param funcSize	(in/out) number of elements in the hash functions
   * @param hashSize	(in/out) number of elements in the hash tables
   * @return		true if the initialization succeeded
   */
  bool init (unsigned numTables,
	     unsigned& funcSize,
	     unsigned& hashSize);

protected:
  /** Add a state to the graph
   * @param state	the encoded state
   * @param size	length of the encoded state, in bytes
   * @return		true if the state was enqueued to mySearch
   */
  bool do_add (const void* state,
	       size_t size);

  /** Get an inflated state from myPathFile
   * @param pos		file offset to the state
   */
  word_t* getState (long pos) const;

public:
  /** Fetch an encoded state
   * @param states	the search list of states
   * @param tail	flag: retrieve from tail of list instead of head
   * @param size	(output) length of the encoded stream (optional)
   * @return		the encoded state
   */
  word_t* pop (bool tail, size_t* size);

private:
  /** Fetch a state from a counterexample path
   * @param u	cursor to the counterexample path
   * @return	the encoded state, inflated
   */
  word_t* getState (class ByteUnpacker& u) const;

  /** Number of hash tables */
  unsigned myNumTables;
  /** Number of elements in the hash function tables (power of 2) */
  unsigned myFuncSize;
  /** Number of elements in the hash tables (power of 2) */
  unsigned myHashSize;
  /** The hash functions */
  size_t** myHashers;
  /** The hash tables */
  size_t** myHashes;
};

#endif // HASHGRAPH_H_
