// Parallelized, lossless reachability set storage -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__

#include "ParSet.h"
#include <sys/types.h>
#if defined WIN32 || defined __WIN32
# undef __STRICT_ANSI__
# include <winsock.h>
typedef long ssize_t;
# undef HAVE_POLL
#else
# include <sys/socket.h>
# if defined __CYGWIN__ || defined __APPLE__
#  undef HAVE_POLL
#  include <sys/select.h>
# else
#  define HAVE_POLL
#  include <sys/poll.h>
# endif
#endif
#ifndef USE_MMAP
# include <unistd.h>
#endif // !USE_MMAP

#ifdef __alpha
# include <sys/errno.h>
#endif // __alpha
#include <cerrno>

#ifndef MSG_NOSIGNAL
# define MSG_NOSIGNAL 0
#endif // MSG_NOSIGNAL

ParSet::ParSet (bool compress, int s) :
  StateSet (compress),
  myFD (s), myRecv (), mySend (), myRecvCount (0), mySendCount (0)
{
}

ParSet::~ParSet ()
{
  close (myFD);
}

/** Send pending data to a socket
 * @param fd	the socket to be written
 * @param sbuf	the buffer to write
 * @param sbufu	index of the first unsent byte in sbuf
 */
static void
sendBuf (int fd,
	 class BytePacker& sbuf,
	 unsigned& sbufu)
{
  assert (sbuf.getLength () > sbufu);
  const char* buf = reinterpret_cast<const char*>(sbuf.getBuf ());
  ssize_t i = send (fd, buf + sbufu, sbuf.getLength () - sbufu,
		    MSG_NOSIGNAL);
  if (i < 0) {
#if defined WIN32 || defined __WIN32
    fprintf (stderr, "send: WinSock error %d\n", WSAGetLastError ());
#else
    perror ("send");
#endif
    exit (-1);
  }
  if ((sbufu += i) == sbuf.getLength ())
    sbuf.clear (), sbufu = 0;
}

void
ParSet::reject (const void* dstate,
		size_t dlen,
		const class Property*,
		const class Net&,
		const class Printer&,
		card_t,
		enum StateSet::Code reason) const
{
  assert (!dstate == !dlen);
  mySend.append (reason);
  mySend.append (dlen);
  mySend.append (dstate, dlen);
  // flush the buffer to notify the server of the rejected state
  do
    sendBuf (myFD, mySend, mySendCount);
  while (mySendCount);
}

bool
ParSet::do_add (const void* buf,
		size_t size)
{
  mySearch.push (buf, size, 0);
  return true;
}

word_t*
ParSet::pop (bool, size_t* size)
{
  unsigned numBytes;
  word_t* buf;
  if (unsigned numStates = mySearch.size () + (mySuccessors != None)) {
    if (const size_t num = mySuppressed.size ()) {
      const struct StateSet::States::state& s = mySuppressed[num - 1];
      if (!getNumArcs ()) {
	assert (mySearch.empty ());
	assert (mySuccessors == One);
	mySend.append (initialState);
	goto append;
      }
      else if (mySuccessors == One) {
	numBytes = s.size;
	buf = new word_t[1 + (numBytes - 1) / sizeof (word_t)];
	memcpy (buf, s.buf, numBytes);
	mySuccessors = None;
	goto fetch;
      }
      else {
	mySend.append (addStates), mySend.append (numStates);
	if (mySuccessors == Many) {
	append:
	  mySend.append (s.size), mySend.append (s.buf, s.size);
	  mySuccessors = None;
	}
      }
    }
    else if (getNumArcs ())
      mySend.append (addStates), mySend.append (numStates);
    else
      assert (numStates == 1), mySend.append (initialState);
    for (StateList::const_iterator i = mySearch.begin ();
	 i != mySearch.end (); i++)
      mySend.append (i->size), mySend.append (i->data, i->size);
  }
  else
    mySend.append (addStates), mySend.append (0);

  assert (mySuccessors == None);
  mySuppressed.clear ();
  mySearch.clear ();
 again:
#ifndef HAVE_POLL
  fd_set readfds, writefds;
  FD_ZERO (&readfds); FD_ZERO (&writefds);
  FD_SET (myFD, &readfds); FD_SET (myFD, &writefds);
  switch (select (myFD + 1, &readfds, &writefds, 0, 0)) {
  default:
#if defined WIN32 || defined __WIN32
    if (WSAGetLastError () == WSAEINTR)
      goto again;
    fprintf (stderr, "select: WinSock error %d\n", WSAGetLastError ());
#else
    if (errno == EINTR)
      goto again;
    perror ("select");
#endif
    return 0;
  case 1:
  case 2:
    if (FD_ISSET (myFD, &writefds))
      sendBuf (myFD, mySend, mySendCount);
    if (FD_ISSET (myFD, &readfds))
      goto enqueue;
    // fall through: no data available for reading
  case 0:
    goto dequeue;
  }
#else
  struct pollfd ufd;
  ufd.fd = myFD;
  ufd.events = POLLIN | POLLOUT;
  ufd.revents = 0;
  switch (poll (&ufd, 1, 0)) {
  default:
    if (errno == EINTR)
      goto again;
    perror ("poll");
    return 0;
  case 1:
    if (ufd.revents & POLLOUT)
      sendBuf (myFD, mySend, mySendCount);
    if (ufd.revents & POLLIN)
      goto enqueue;
    // fall through: no data available for reading
  case 0:
    goto dequeue;
  }
#endif

 dequeue:
  {
    class ByteUnpacker u (myRecv.getBuf () + myRecvCount);
    if (!u.extract (myRecv, numBytes) ||
	!u.ensureData (myRecv, numBytes)) {
      assert (myRecv.getBuf () + myRecvCount <= u.buf);
      goto enqueue;
    }
    if ((myRecvCount = u.buf - myRecv.getBuf () + numBytes) ==
	myRecv.getLength ())
      myRecvCount = 0, myRecv.clear ();

    buf = new word_t[1 + (numBytes - 1) / sizeof (word_t)];
    memcpy (buf, u.buf, numBytes);
  }

  if (compress && !mySuppressed.add (buf, numBytes))
    assert (false);

 fetch:
  if (size)
    *size = numBytes;
  else
    BitPacker::inflate (buf[(numBytes - 1) / sizeof (word_t)],
			(-numBytes) % sizeof (word_t));
  return buf;

 enqueue:
  assert (myRecvCount
	  ? myRecvCount < myRecv.getLength ()
	  : !myRecv.getLength ());
  myRecv.allocate (2048);

  ssize_t rlen = recv (myFD, reinterpret_cast<char*>
		       (myRecv.getBuf () + myRecv.getLength ()),
		       myRecv.getAllocated () - myRecv.getLength (),
		       MSG_NOSIGNAL);
  if (rlen < 0) {
#if defined WIN32 || defined __WIN32
    int wsaerror = WSAGetLastError ();
    if (wsaerror != WSAECONNRESET)
      fprintf (stderr, "recv: WinSock error %d\n", wsaerror);
#else
    if (errno != ECONNRESET)
      perror ("recv");
#endif
    return 0;
  }
  else if (!rlen) // the remote end tells us to terminate
    return 0;

  myRecv.setLength (myRecv.getLength () + rlen);
  goto dequeue;
}
