// Storage of encoded states -*- c++ -*-

/** @file StateSet.C
 * Set of reached states, represented with bit vectors
 */

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "StateSet.h"
#include "Printer.h"
#include "Net.h"
#include "Dotty.h"
#include "ByteBuffer.h"

bool
StateSet::openFile ()
{
  if (!(myPathFile = tmpfile ()))
    perror ("tmpfile");
  return !!myPathFile;
}

void
StateSet::reject (const void* dstate,
		  size_t dlen,
		  const class Property* prop,
		  const class Net& net,
		  const class Printer& printer,
		  card_t visual,
		  enum Code reason) const
{
  const class Dotty* dotty;
  if (visual && (dotty = Dotty::startVisual ()))
    dotty->displayPrologue (visual > 1);
  else
    dotty = 0;

  switch (reason) {
  case initialState:
  case addStates:
    assert (false);
    break;
  case deadlockState:
    printer.printRaw ("deadlock");
    break;
  case deadlockFatal:
    printer.printRaw ("fatal deadlock");
    break;
  case rejectState:
    printer.printRaw ("rejected state");
    break;
  case rejectFatal:
    printer.printRaw ("fatally rejected state");
    break;
  case propertyError:
    printer.printRaw ("property evaluation error");
    break;
  case inconsistent:
    printer.printRaw ("inconsistent successor");
    break;
  }

  printer.finish ();

  if (!myPathFile) {
    if (dstate)
      net.displayArcs (printer, dstate, 0, prop, dotty, false);
    return;
  }

  unsigned numOffsets = 1;
  unsigned* offsets = new unsigned[1];
  assert (myOffset < ftell (myPathFile));
  if ((*offsets = myOffset)) {
    unsigned char rbuf[8];
    class ByteUnpacker u (rbuf);
    do {
      fseek (myPathFile, offsets[numOffsets - 1], SEEK_SET);
      if (!(numOffsets & (numOffsets + 1))) {
	unsigned* o = new unsigned[(numOffsets + 1) << 1];
	memcpy (o, offsets, numOffsets * sizeof *offsets);
	delete[] offsets;
	offsets = o;
      }
      fread (rbuf, sizeof rbuf, 1, myPathFile);
      offsets[numOffsets] = u.extract (); u.buf = rbuf;
    }
    while (offsets[numOffsets++]);
  }

  word_t* s1 = 0;
  for (const unsigned* o = offsets + numOffsets; o-- > offsets; ) {
    word_t* state = getState (*o);
    if (!s1)
      net.displayArcs (printer, state, 0, prop, dotty, compress);
    else
      net.displayArcs (printer, s1, state, prop, dotty, compress);
    delete[] s1;
    s1 = state;
    /** Flag: has the analysis been interrupted? */
    extern volatile bool interrupted;
    if (interrupted)
      goto intr;
  }

  if (dstate) {
    word_t* d1 = new word_t[(dlen + (sizeof (word_t) - 1)) / sizeof (word_t)];
    memcpy (d1, dstate, dlen);
    BitPacker::inflate (d1[(dlen - 1) / sizeof (word_t)],
			(-dlen) % sizeof (word_t));
    net.displayArcs (printer, s1, d1, prop, dotty, compress);
    delete[] d1;
  }

 intr:
  delete[] s1;
  delete[] offsets;

  fseek (myPathFile, 0, SEEK_END);

  if (dotty)
    dotty->displayEpilogue ();
}

void
StateSet::showProgress (const class Printer& printer,
			unsigned pending) const
{
  if (Net::theEdgeThresold && !(getNumArcs () % Net::theEdgeThresold)) {
    if (getNumStates ())
      printer.print (getNumStates ()), printer.printRaw (" states, ");
    printer.print (getNumArcs ()), printer.printRaw (" arcs, ");
    printer.print (card_t (mySearch.size ()));
    if (pending)
      printer.delimiter ('+'), printer.print (pending);
    printer.printRaw (" pending");
    printer.finish ();
  }
}

word_t*
StateSet::getState (long) const
{
  assert (false);
  return 0;
}
