// Storage of encoded states -*- c++ -*-

#ifndef STATESET_H_
# define STATESET_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include "StateList.h"
# include <cstdio>

/** @file StateSet.h
 * Set of reached states, represented with bit vectors
 */

/** Set of reached states */
class StateSet
{
public:
  /** codes reported from worker to server */
  enum Code {
    initialState,	//< report the initial state
    addStates,		//< add successor states
    deadlockState,	//< notify of a rejected deadlock state
    deadlockFatal,	//< notify of a fatally rejected deadlock
    rejectState,	//< notify of a rejected state
    rejectFatal,	//< notify of a fatally rejected state
    propertyError,	//< notify of property evaluation error
    inconsistent	//< notify of an inconsistent successor state
  };
  /** Collection of suppressed states in path compression */
  class States {
  public:
    /** An encoded state */
    struct state {
      size_t size;	//< length of the encoded state, in bytes
      word_t* buf;	//< the deflated encoded state
    };
    /** Constructor */
    States () : myStates (0), myNumber (0), myAllocated (0) {}
  private:
    /** Copy constructor */
    States (const class Suppressed& old);
    /** Assignemnt operator */
    class States& operator= (const class States& other);
  public:
    /** Destructor */
    ~States () {
      for (unsigned i = myNumber; i--; )
	delete[] myStates[i].buf;
      operator delete (myStates);
    }

    /** Get the number of states in the set */
    size_t size () const { return myNumber; }

    /** Index the state array */
    const struct state& operator[] (unsigned i) {
      return myStates[i];
    }

    /** Clear the set */
    void clear () {
      for (unsigned i = myNumber; i--; )
	delete[] myStates[i].buf;
      memset (myStates, 0, myNumber * sizeof *myStates);
      myNumber = 0;
    }

    /** Search for a state in the set
     * @param buf	the deflated encoded state
     * @param size	length of the state in bytes
     * @return		the index of the state + 1, or 0 if not found
     */
    unsigned search (const void* buf,
		     size_t size) const {
      unsigned i = myNumber;
      while (i--)
	if (size == myStates[i].size &&
	    !memcmp (myStates[i].buf, buf, size))
	  break;
      return i + 1;
    }

    /** Add a state to the set
     * @param buf	the deflated encoded state
     * @param size	length of the state in bytes
     * @return		true if the state was new
     */
    bool add (const void* buf,
	      size_t size) {
      if (search (buf, size))
	return false;
      if (!myAllocated) {
	assert (!myNumber && !myStates);
	myAllocated = 1;
	myStates = reinterpret_cast<struct state*>
	  (operator new (sizeof *myStates));
      }
      else if (myNumber >= myAllocated) {
	assert (myNumber && myStates);
	struct state* states = reinterpret_cast<struct state*>
	  (memcpy (operator new ((myAllocated << 1) * sizeof *states),
		   myStates, myAllocated * sizeof *states));
	delete[] myStates;
	myStates = states;
	myAllocated <<= 1;
      }

      struct state& s = myStates[myNumber++];
      s.size = size;
      memcpy (s.buf = new word_t[1 + (size - 1) / sizeof (word_t)],
	      buf, size);
      return true;
    }

# if !defined __sgi || defined __GNUC__
  private:
# endif // MIPS CC bug work-around
    struct state* myStates;
  private:
    /** Number of used entries in myStates */
    unsigned myNumber;
    /** Number of entries allocated for myStates */
    unsigned myAllocated;
  };

  /** Constructor
   * @param compress	flag: apply path compression
   */
  StateSet (bool compress) :
    compress (compress),
    myNumStates (0), myNumArcs (0), myMinSize (size_t (-1)), myMaxSize (0),
    mySearch (),
    myPathFile (0), myOffset (0), mySuccessors (None), mySuppressed () {}
private:
  /** Copy constructor */
  StateSet (const class StateSet& old);
  /** Assignment operator */
  class StateSet& operator= (const class StateSet& old);
public:
  /** Destructor */
  virtual ~StateSet () {
    if (myPathFile) fclose (myPathFile);
  }

  /** Get the number of states */
  unsigned getNumStates () const { return myNumStates; }
  /** Get the number of arcs (number of add () invocations) */
  unsigned getNumArcs () const { return myNumArcs - 1; }
  /** Get the minimum length of state vectors */
  size_t getMinSize () const { return myMinSize; }
  /** Get the maximum length of state vectors */
  size_t getMaxSize () const { return myMaxSize; }

  /** Open the counterexample path file
   * @return		true on success; false on failure
   */
  bool openFile ();

  /** Display path to a rejected state
   * @param dstate	the encoded rejected state (0=deadlock)
   * @param dlen	length of dstate in bytes
   * @param prop	a safety property to be checked (optional)
   * @param net		the net
   * @param printer	the output stream for counterexample
   * @param visual	visualization level
   * @param reason	reason for rejecting the state
   */
  virtual void reject (const void* dstate,
		       size_t dlen,
		       const class Property* prop,
		       const class Net& net,
		       const class Printer& printer,
		       card_t visual,
		       enum Code reason) const;
 
  /** Add a state to the set
   * @param state	the encoded state
   * @param size	length of the encoded state, in bytes
   * @return		true if the state was enqueued to mySearch
   */
  bool add (const void* state,
	    size_t size) {
    myNumArcs++;
    if (size > myMaxSize) myMaxSize = size;
    if (size < myMinSize) myMinSize = size;
    if (compress) {
      if (!mySuccessors && mySuppressed.add (state, size))
	return mySuccessors = One, false;
      mySuccessors = Many;
    }
    return do_add (state, size);
  }

  /** Display the number of processed states and arcs
   * @param printer	the output stream
   * @param pending	number of additionally pending states
   */
  void showProgress (const class Printer& printer,
		     unsigned pending = 0) const;

protected:
  /** Increment the state count */
  void newState () { myNumStates++; }

  /** Add a state to the set
   * @param state	the encoded state
   * @param size	length of the encoded state, in bytes
   * @return		true if the state was enqueued to mySearch
   */
  virtual bool do_add (const void* state,
		       size_t size) = 0;

  /** Get an inflated state from myPathFile
   * @param pos		file offset to the state
   */
  virtual word_t* getState (long pos) const;

public:
  /** Fetch an encoded state
   * @param tail	flag: retrieve from tail of list instead of head
   * @param size	(output) length of the encoded stream (optional)
   * @return		the encoded state
   */
  virtual word_t* pop (bool tail, size_t* size = 0) = 0;

  /** Set the offset to the counterexample path file */
  void setOffset (long offset) { myOffset = offset; }
  /** Read the offset to the counterexample path file */
  long getOffset () const { return myOffset; }

public:
  /** flag: apply path compression */
  const bool compress;
private:
  /** Number of generated states */
  unsigned myNumStates;
  /** Number of generated arcs (invocations to add ()) */
  unsigned myNumArcs;
  /** Minimum length of encoded state vectors */
  size_t myMinSize;
  /** Maximum length of encoded state vectors */
  size_t myMaxSize;
protected:
  /** The search list of states */
  class StateList mySearch;
  /** Counterexample path file */
  FILE* myPathFile;
  /** Current state offset in the counterexample path file */
  long myOffset;
  /** Flag: how many successors does the states have (when compress==true) */
  enum { None = 0, One, Many } mySuccessors;
# if defined __sgi && !defined __GNUC__
public:
# endif // MIPS CC bug work-around
  /** Suppressed states (when compress==true) */
  class States mySuppressed;
};

#endif // STATESET_H_
