// Transition analysis routines -*- c++ -*-
// #included from Transition.C for each variation

#ifdef STATE_DECL
/** Check for unacceptable valuations and clear the errors
 * @param status	default return value
 * @param valuation	valuation to be checked
 * @param net		the net that is being analyzed
 * @param STATE_DECL	parameters that identify the state
 * @param transition	transition name
 * @param printer	printer object for diagnostic output
 * @param reason	reason for failure (optional)
 * @return		status, or false if there is an error in the valuation
 */
static bool
willDo (bool status,
	class Valuation& valuation,
	const class Net& net,
	STATE_DECL,
	const char* transition,
	const class Printer& printer,
	const char* reason = 0)
{
  if (!valuation.isOKorVar () || reason) {
    status = false;
    if (valuation.getError () == errFatal)
      net.flagFatal ();
    if (displayState (net, STATE_PARAM, printer)) {
      if (transition)
	printer.printQuoted (transition);
      else
	printer.printRaw ("(anonymous)");
      printer.delimiter (':');
      if (reason) printer.printRaw (reason);
      printer.linebreak ();
      valuation.display (printer);
      printer.finish ();
    }
  }

  valuation.clearErrors ();
  return status;
}

/** Augment the valuation with initial assignments to output variables
 * @param transition	the transition
 * @param valuation	the valuation
 * @param net		the net that is being analyzed
 * @param STATE_DECL	parameters that identify the state
 * @param printer	printer object for diagnostic output
 * @return		true if the operation succeeded
 */
static bool
firstOutput (const class Transition& transition,
	     class Valuation& valuation,
	     const class Net& net,
	     STATE_DECL,
	     const class Printer& printer)
{
  if (const class QuantifierList* vars = transition.getOutputVariables ()) {
    switch (vars->augment (valuation)) {
    case Quantifier::pass:
      return true;
    case Quantifier::fail:
      switch (vars->advance (valuation)) {
      case Quantifier::pass:
	return true;
      case Quantifier::fail:
	willDo (false, valuation, net, STATE_PARAM,
		transition.getName (), printer, "empty output:");
	vars->clear (valuation);
	return false;
      case Quantifier::undefined:
	break;
      }
      // fall through
    case Quantifier::undefined:
      willDo (false, valuation, net, STATE_PARAM,
	      transition.getName (), printer, "first output:");
      vars->clear (valuation);
      return false;
    }
  }
  return true;
}

/** Iterate to the next assignment to output variables
 * @param transition	the transition
 * @param valuation	the valuation
 * @param net		the net that is being analyzed
 * @param STATE_DECL	parameters that identify the state
 * @param printer	printer object for diagnostic output
 * @return		true if a new valuation was generated
 */
static bool
nextOutput (const class Transition& transition,
	    class Valuation& valuation,
	    const class Net& net,
	    STATE_DECL,
	    const class Printer& printer)
{
  if (const class QuantifierList* vars = transition.getOutputVariables ()) {
    switch (vars->advance (valuation)) {
    case Quantifier::pass:
      return true;
    case Quantifier::undefined:
      willDo (false, valuation, net, STATE_PARAM,
	      transition.getName (), printer, "iterating output:");
      // fall through
    case Quantifier::fail:
      vars->clear (valuation);
      break;
    }
  }
  return false;
}
#endif // STATE_DECL

/** Fire the output arcs of a transition
 * @param transition	the transition
 * @param m		the state (input/output parameter)
 * @param valuation	the transition instance
 * @param net		the net the transition belongs to
 * @param STATE_DECL2	parameters that identify the state storage
 * @param printer	printer object for diagnostic output
 * @return		true if any instances of the transition were fired
 */
static bool
fireOutputs (const class Transition& transition,
	     class GlobalMarking& m,
	     class Valuation& valuation,
	     const class Net& net,
	     STATE_DECL2,
	     const class Printer& printer)
{
  assert (&net.getTransition (transition.getIndex ()) == &transition &&
	  valuation.isOK ());

  for (unsigned out = 0; out < transition.getNumOutputs (); out++) {
    const class Arc& arc = transition.getOutput (out);
    class PlaceMarking& place = m[arc.getPlace ()];
    if (!arc.getExpr ().add (place, 1, valuation)) {
      willDo (false, valuation, net, STATE_PARAM2,
	      transition.getName (), printer, "firing:");
      return false;
    }
  }

  card_t errorplace;
STATE_ADD1
  if (STATE_ADD_FAIL) {
    willDo (false, valuation, net, STATE_PARAM2,
	    transition.getName (), printer, "encoding successor:");
    printer.printRaw ("place ");
    printer.print (net.getPlace (errorplace).getName ());
    printer.printRaw (": integrity violation");
    printer.finish ();
  }
  else {
STATE_ADD2
  }

  return true;
}

bool
Transition::analyze (class GlobalMarking& m,
		     const class Net& net,
		     STATE_DECL2,
		     const class Printer& printer) const
{
  if (!myGates.empty () &&
      (*myGates.begin ())->getKind () == Expression::eConstant) {
    // there is a constant gate (which should be always disabled):
    // ignore the transition
#ifndef NDEBUG
    const class Expression* g = *myGates.begin ();
    const class Value& v = static_cast<const class Constant*>(g)->getValue ();
    assert (v.getKind () == Value::vLeaf);
    assert (!bool (static_cast<const class LeafValue&>(v)));
#endif // NDEBUG
    return false;
  }

  /** Flag: were any successors generated? */
  bool succ = false;
  class Valuation valuation;
  const struct unif* u;
  VariableSet::const_iterator vi;
  /** processed tokens */
  class TokenList tokens;

  for (class Token* token = 0;;) {
    assert (valuation.isOK ());
    if (token && (u = &token->getUnifier ())->vars) {
    unify:
#ifndef NDEBUG
      for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	assert (!valuation.getValue (**vi) || (*vi)->isUndefined ());
#endif // NDEBUG

      if (!token->getConcrete ()) /* All value candidates have been checked. */
	goto cont2;
      if (!willDo (token->isCompatible (valuation), valuation, net,
		   STATE_PARAM2, myName, printer))
	goto incompatible;

      u->m->getToken ()->getLvalues (token->getValue (), valuation, *u->vars);
#ifndef NDEBUG
      for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	assert (!!valuation.getValue (**vi));
#endif // NDEBUG
      if (!willDo (checkGates (valuation) &&
		   token->isCompatible (valuation) &&
		   tokens.isCompatible (valuation),
		   valuation, net, STATE_PARAM2, myName, printer)) {
	if (u->varMult) {
	  for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	    if (!token->isUnified (**vi)) valuation.undefine (**vi);
	}
	else
	  for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	    valuation.undefine (**vi);
	goto incompatible;
      }

    tokenBinding:
      token->reserve ();
      tokens.push (*token);
    }

    assert (valuation.isOK ());
    if (token)
      u = token->getUnifier ().next, token = 0;
    else if (!tokens.empty ())
      u = (*tokens.begin ())->getUnifier ().next;
    else
      u = myUnif;

  next:
    if (u) {
      card_t mult = u->m->getMultiplicity (valuation);
      if (valuation.isOK ()) {
	if (!mult) {
	  u = u->next;
	  goto next;
	}
	else if (m[u->place].empty ())
	  goto cont;
	token = new class Token (mult, m[u->place], *u);
	if (u->vars) {
	  if (u->varMult)
	    for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	      if (valuation.getValue (**vi)) token->addUnified (**vi);
	  goto unify;
	}
	else if (token->isBindable (valuation)) {
	  assert (valuation.isOK ());
	  goto tokenBinding;
	}
	if (valuation.isOK ())
	  goto cont;
      }
      willDo (false, valuation, net, STATE_PARAM2, myName, printer,
	      "unification:");
    }
    else if (willDo (true, valuation, net, STATE_PARAM2, myName, printer,
		     checkGates (valuation, errNone)
		     ? 0 : "undefined gate:") &&
	     firstOutput (*this, valuation, net, STATE_PARAM2, printer)) {
      /** tokens removed from the input places */
      class GlobalMarking* inputs = 0;
      if (myOutputMarking) {
	inputs = new class GlobalMarking (net);
	for (TokenList::const_iterator t = tokens.begin ();
	     t != tokens.end (); t++)
	  if (!(*t)->copyRemoved (*inputs))
	    assert (false);
      }
      valuation.setGlobalMarking (inputs);

      do {
	class GlobalMarking successor (m);
	if (fireOutputs (*this, successor, valuation, net,
			 STATE_PARAM3, printer)) {
	  succ = true;
#ifdef STATE_SUCC
	  STATE_SUCC;
#endif // STATE_SUCC
	}
      } while (!interrupted && !net.isFatal () &&
	       nextOutput (*this, valuation, net, STATE_PARAM2, printer));

      valuation.setGlobalMarking (0);
      delete inputs;
    }

  cont:
    if (token) {
      if ((u = &token->getUnifier ())->vars)
	if (u->varMult) {
	  for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	    if (!token->isUnified (**vi)) valuation.undefine (**vi);
	}
	else
	  for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	    valuation.undefine (**vi);
    cont2:
      delete token;
    }

    if (tokens.empty ()) /* all bindings have been checked */
      break;

    token = &tokens.pop ();

    if (token->isReserved ())
      token->free ();

    if ((u = &token->getUnifier ())->vars) {
      if (u->varMult) {
	for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	  if (!token->isUnified (**vi)) valuation.undefine (**vi);
      }
      else
	for (vi = u->vars->begin (); vi != u->vars->end (); vi++)
	  valuation.undefine (**vi);
    incompatible:
      if (interrupted || net.isFatal ())
	break;
      if (token->next ())
	goto unify;
      goto cont2;
    }
    else
      goto cont;
  }

  return succ;
}

#undef STATE_DECL
#undef STATE_PARAM
#undef STATE_DECL2
#undef STATE_PARAM2
#undef STATE_PARAM3
#undef STATE_ADD1
#undef STATE_ADD_FAIL
#undef STATE_ADD2
#undef STATE_ADD3
#undef STATE_SUCC
